@@ -30,7 +30,7 @@ impl ChatMassage {
         ChatMassage {
             kind: message.kind,
             from: message.from.clone().unwrap_or_default(),
-            text: message.text.clone(),
+            text: message.plain_text(),
             tick,
         }
     }
@@ -103,7 +103,7 @@ pub struct ClassList([u8; 10]);
 
 impl ClassList {
     /// Get an iterator for all classes played and the number of spawn on the class
-    pub fn iter<'a>(&'a self) -> impl Iterator<Item = (Class, u8)> + 'a {
+    pub fn iter(&self) -> impl Iterator<Item = (Class, u8)> + '_ {
         self.0
             .iter()
             .copied()
@@ -224,9 +224,9 @@ impl From<crate::demo::data::UserInfo> for UserInfo {
     fn from(info: crate::demo::data::UserInfo) -> Self {
         UserInfo {
             classes: ClassList::default(),
-            name: info.name,
-            user_id: info.user_id,
-            steam_id: info.steam_id,
+            name: info.player_info.name,
+            user_id: info.player_info.user_id.into(),
+            steam_id: info.player_info.steam_id,
             entity_id: info.entity_id,
             team: Team::default(),
         }
@@ -302,10 +302,10 @@ impl MessageHandler for Analyser {
     type Output = MatchState;
 
     fn does_handle(message_type: MessageType) -> bool {
-        match message_type {
-            MessageType::GameEvent | MessageType::UserMessage | MessageType::ServerInfo => true,
-            _ => false,
-        }
+        matches!(
+            message_type,
+            MessageType::GameEvent | MessageType::UserMessage | MessageType::ServerInfo
+        )
     }
 
     fn handle_message(&mut self, message: &Message, tick: u32) {
@@ -323,14 +323,11 @@ impl MessageHandler for Analyser {
     }
 
     fn handle_string_entry(&mut self, table: &str, _index: usize, entry: &StringTableEntry) {
-        match table {
-            "userinfo" => {
-                let _ = self.parse_user_info(
-                    entry.text.as_ref().map(|s| s.as_ref()),
-                    entry.extra_data.as_ref().map(|data| data.data.clone()),
-                );
-            }
-            _ => {}
+        if table == "userinfo" {
+            let _ = self.parse_user_info(
+                entry.text.as_ref().map(|s| s.as_ref()),
+                entry.extra_data.as_ref().map(|data| data.data.clone()),
+            );
         }
     }
 
@@ -395,7 +392,7 @@ impl Analyser {
         if let Some(user_info) = crate::demo::data::UserInfo::parse_from_string_table(text, data)? {
             self.state
                 .users
-                .entry(user_info.user_id)
+                .entry(user_info.player_info.user_id.into())
                 .and_modify(|info| {
                     info.entity_id = user_info.entity_id;
                 })
