@@ -1,4 +1,4 @@
-use bitbuffer::{BitRead, LittleEndian};
+use bitbuffer::{BitRead, BitWrite, BitWriteSized, BitWriteStream, LittleEndian};
 use enumflags2::{bitflags, BitFlags};
 use serde::{Deserialize, Serialize};
 
@@ -10,17 +10,29 @@ use crate::consthash::ConstFnvHash;
 use crate::demo::message::stringtable::log_base2;
 use crate::demo::packet::datatable::SendTableName;
 use crate::demo::parser::MalformedSendPropDefinitionError;
+use num_traits::Signed;
 use parse_display::Display;
 use std::cmp::min;
 use std::convert::{TryFrom, TryInto};
 use std::fmt;
 use std::hash::Hash;
-use std::rc::Rc;
+use std::ops::BitOr;
 
 #[derive(
-    BitRead, PartialEq, Eq, Hash, Debug, Display, Clone, Serialize, Deserialize, Ord, PartialOrd,
+    BitRead,
+    BitWrite,
+    PartialEq,
+    Eq,
+    Hash,
+    Debug,
+    Display,
+    Clone,
+    Serialize,
+    Deserialize,
+    Ord,
+    PartialOrd,
 )]
-pub struct SendPropName(Rc<String>);
+pub struct SendPropName(String);
 
 impl SendPropName {
     pub fn as_str(&self) -> &str {
@@ -36,7 +48,13 @@ impl PartialEq<&str> for SendPropName {
 
 impl From<String> for SendPropName {
     fn from(value: String) -> Self {
-        Self(Rc::new(value))
+        Self(value)
+    }
+}
+
+impl From<&str> for SendPropName {
+    fn from(value: &str) -> Self {
+        value.to_string().into()
     }
 }
 
@@ -44,7 +62,7 @@ impl From<String> for SendPropName {
 pub struct RawSendPropDefinition {
     pub prop_type: SendPropType,
     pub name: SendPropName,
-    pub owner_table: SendTableName,
+    pub identifier: SendPropIdentifier,
     pub flags: SendPropFlags,
     pub table_name: Option<SendTableName>,
     pub low_value: Option<f32>,
@@ -52,6 +70,7 @@ pub struct RawSendPropDefinition {
     pub bit_count: Option<u32>,
     pub element_count: Option<u16>,
     pub array_property: Option<Box<RawSendPropDefinition>>,
+    pub original_bit_count: Option<u32>,
 }
 
 impl PartialEq for RawSendPropDefinition {
@@ -65,8 +84,7 @@ impl fmt::Display for RawSendPropDefinition {
         match self.prop_type {
             SendPropType::Vector | SendPropType::VectorXY => write!(
                 f,
-                "{}::{}({})(flags: {}, low: {}, high: {}, bits: {})",
-                self.owner_table,
+                "{}({})(flags: {}, low: {}, high: {}, bits: {})",
                 self.name,
                 self.prop_type,
                 self.flags,
@@ -76,8 +94,7 @@ impl fmt::Display for RawSendPropDefinition {
             ),
             SendPropType::Float => write!(
                 f,
-                "{}::{}({})(flags: {}, low: {}, high: {}, bits: {})",
-                self.owner_table,
+                "{}({})(flags: {}, low: {}, high: {}, bits: {})",
                 self.name,
                 self.prop_type,
                 self.flags,
@@ -87,21 +104,19 @@ impl fmt::Display for RawSendPropDefinition {
             ),
             SendPropType::Int => write!(
                 f,
-                "{}::{}({})(flags: {}, bits: {})",
-                self.owner_table,
+                "{}({})(flags: {}, bits: {})",
                 self.name,
                 self.prop_type,
                 self.flags,
                 self.bit_count.unwrap_or(32)
             ),
             SendPropType::String => {
-                write!(f, "{}::{}({})", self.owner_table, self.name, self.prop_type)
+                write!(f, "{}({})", self.name, self.prop_type)
             }
             SendPropType::Array => match &self.array_property {
                 Some(array_prop) => write!(
                     f,
-                    "{}::{}([{}({})] * {})",
-                    self.owner_table,
+                    "{}([{}({})] * {})",
                     self.name,
                     array_prop.prop_type,
                     array_prop.flags,
@@ -110,15 +125,11 @@ impl fmt::Display for RawSendPropDefinition {
                 None => write!(f, "{}(Malformed array)", self.name),
             },
             SendPropType::DataTable => match &self.table_name {
-                Some(sub_table) => write!(
-                    f,
-                    "{}::{}(DataTable = {})",
-                    self.owner_table, self.name, sub_table
-                ),
+                Some(sub_table) => write!(f, "{}(DataTable = {})", self.name, sub_table),
                 None => write!(f, "{}(Malformed DataTable)", self.name),
             },
             SendPropType::NumSendPropTypes => {
-                write!(f, "{}::{}(NumSendPropTypes)", self.owner_table, self.name)
+                write!(f, "{}(NumSendPropTypes)", self.name)
             }
         }
     }
@@ -126,13 +137,13 @@ impl fmt::Display for RawSendPropDefinition {
 
 impl RawSendPropDefinition {
     pub fn identifier(&self) -> SendPropIdentifier {
-        SendPropIdentifier::new(self.owner_table.as_str(), self.name.as_str())
+        self.identifier
     }
 
     pub fn with_array_property(self, array_property: Self) -> Self {
         RawSendPropDefinition {
             prop_type: self.prop_type,
-            owner_table: self.owner_table,
+            identifier: self.identifier,
             name: self.name,
             flags: self.flags,
             table_name: self.table_name,
@@ -141,6 +152,7 @@ impl RawSendPropDefinition {
             bit_count: self.bit_count,
             element_count: self.element_count,
             array_property: Some(Box::new(array_property)),
+            original_bit_count: self.original_bit_count,
         }
     }
 
@@ -157,9 +169,10 @@ impl RawSendPropDefinition {
         }
     }
 
-    pub fn read(stream: &mut Stream, owner_table: SendTableName) -> ReadResult<Self> {
+    pub fn read(stream: &mut Stream, owner_table: &SendTableName) -> ReadResult<Self> {
         let prop_type = SendPropType::read(stream)?;
-        let name = stream.read_string(None)?.to_string().into();
+        let name: SendPropName = stream.read()?;
+        let identifier = SendPropIdentifier::new(owner_table.as_str(), name.as_str());
         let flags = SendPropFlags::read(stream)?;
         let mut table_name = None;
         let mut element_count = None;
@@ -175,6 +188,7 @@ impl RawSendPropDefinition {
             high_value = Some(stream.read()?);
             bit_count = Some(stream.read_int(7)?);
         }
+        let original_bit_count = bit_count;
 
         if flags.contains(SendPropFlag::NoScale) {
             if prop_type == SendPropType::Float {
@@ -189,13 +203,14 @@ impl RawSendPropDefinition {
         Ok(RawSendPropDefinition {
             prop_type,
             name,
-            owner_table,
+            identifier,
             flags,
             table_name,
             low_value,
             high_value,
             bit_count,
             element_count,
+            original_bit_count,
             array_property: None,
         })
     }
@@ -213,7 +228,31 @@ impl RawSendPropDefinition {
     }
 }
 
-#[derive(BitRead, Copy, Clone, PartialEq, Debug, Display)]
+impl BitWrite<LittleEndian> for RawSendPropDefinition {
+    fn write(&self, stream: &mut BitWriteStream<LittleEndian>) -> ReadResult<()> {
+        self.prop_type.write(stream)?;
+        self.name.write(stream)?;
+        self.flags.write(stream)?;
+
+        if let Some(table_name) = self.table_name.as_ref() {
+            table_name.write(stream)?;
+        }
+        if let Some(element_count) = self.element_count {
+            element_count.write_sized(stream, 10)?;
+        }
+        if let (Some(low_value), Some(high_value), Some(bit_count)) =
+            (self.low_value, self.high_value, self.original_bit_count)
+        {
+            low_value.write(stream)?;
+            high_value.write(stream)?;
+            bit_count.write_sized(stream, 7)?;
+        }
+
+        Ok(())
+    }
+}
+
+#[derive(BitRead, BitWrite, Copy, Clone, PartialEq, Debug, Display)]
 #[discriminant_bits = 5]
 pub enum SendPropType {
     Int = 0,
@@ -271,9 +310,17 @@ pub enum SendPropFlag {
     NormalVarInt = 32,
 }
 
-#[derive(Debug, Copy, Clone, PartialEq)]
+#[derive(Debug, Copy, Clone, PartialEq, Default)]
 pub struct SendPropFlags(BitFlags<SendPropFlag>);
 
+impl BitOr<SendPropFlag> for SendPropFlags {
+    type Output = SendPropFlags;
+
+    fn bitor(self, rhs: SendPropFlag) -> Self::Output {
+        Self(self.0 | rhs)
+    }
+}
+
 impl fmt::Display for SendPropFlags {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         let debug = format!("{:?}", self.0);
@@ -303,6 +350,12 @@ impl BitRead<'_, LittleEndian> for SendPropFlags {
     }
 }
 
+impl BitWrite<LittleEndian> for SendPropFlags {
+    fn write(&self, stream: &mut BitWriteStream<LittleEndian>) -> ReadResult<()> {
+        self.0.bits().write(stream)
+    }
+}
+
 #[derive(Debug, Clone)]
 pub enum FloatDefinition {
     Coord,
@@ -569,6 +622,11 @@ impl fmt::Display for SendPropValue {
     }
 }
 
+fn float_scale(bit_count: u8) -> f32 {
+    // is this -1 correct?, it is consistent with the js version but seems weird
+    (1i32.wrapping_shl(bit_count as u32)) as f32 - 1.0
+}
+
 impl SendPropValue {
     pub fn parse(stream: &mut Stream, definition: &SendPropParseDefinition) -> Result<Self> {
         match definition {
@@ -629,6 +687,75 @@ impl SendPropValue {
             }
         }
     }
+    pub fn encode(
+        &self,
+        stream: &mut BitWriteStream<LittleEndian>,
+        definition: &SendPropParseDefinition,
+    ) -> Result<()> {
+        match definition {
+            SendPropParseDefinition::NormalVarInt { unsigned, .. } => {
+                let val: i64 = self.try_into()?;
+                write_var_int(val as i32, stream, !*unsigned)?;
+                Ok(())
+            }
+            SendPropParseDefinition::UnsignedInt { bit_count, .. } => {
+                let val: i64 = self.try_into()?;
+                (val as u32).write_sized(stream, *bit_count as usize)?;
+                Ok(())
+            }
+            SendPropParseDefinition::Int { bit_count, .. } => {
+                let val: i64 = self.try_into()?;
+                (val as i32).write_sized(stream, *bit_count as usize)?;
+                Ok(())
+            }
+            SendPropParseDefinition::Float {
+                definition: float_definition,
+                ..
+            } => {
+                let val: f32 = self.try_into()?;
+                Self::write_float(val, stream, float_definition)
+            }
+            SendPropParseDefinition::String { .. } => {
+                let val: &str = self.try_into()?;
+                (val.len() as u16).write_sized(stream, 9)?;
+                val.write_sized(stream, val.len())?;
+                Ok(())
+            }
+            SendPropParseDefinition::Vector {
+                definition: float_definition,
+                ..
+            } => {
+                let val: Vector = self.try_into()?;
+                Self::write_float(val.x, stream, float_definition)?;
+                Self::write_float(val.y, stream, float_definition)?;
+                Self::write_float(val.z, stream, float_definition)?;
+                Ok(())
+            }
+            SendPropParseDefinition::VectorXY {
+                definition: float_definition,
+                ..
+            } => {
+                let val: VectorXY = self.try_into()?;
+                Self::write_float(val.x, stream, float_definition)?;
+                Self::write_float(val.y, stream, float_definition)?;
+                Ok(())
+            }
+            SendPropParseDefinition::Array {
+                count_bit_count,
+                inner_definition,
+                ..
+            } => {
+                let array: &[SendPropValue] = self.try_into()?;
+                (array.len() as u16).write_sized(stream, *count_bit_count as usize)?;
+
+                for inner in array {
+                    inner.encode(stream, inner_definition)?
+                }
+
+                Ok(())
+            }
+        }
+    }
 
     fn read_float(stream: &mut Stream, definition: &FloatDefinition) -> Result<f32> {
         match definition {
@@ -650,13 +777,193 @@ impl SendPropValue {
                 high,
             } => {
                 let raw: u32 = stream.read_int(*bit_count as usize)?;
-                // is this -1 correct?, it is consistent with the js version but seems weird
-                let percentage =
-                    (raw as f32) / ((1i32.wrapping_shl(*bit_count as u32)) as f32 - 1.0);
+                let scale = float_scale(*bit_count);
+                let percentage = (raw as f32) / scale;
                 Ok(low + ((high - low) * percentage))
             }
         }
     }
+
+    fn write_float(
+        val: f32,
+        stream: &mut BitWriteStream<LittleEndian>,
+        definition: &FloatDefinition,
+    ) -> Result<()> {
+        match definition {
+            FloatDefinition::Coord => write_bit_coord(val, stream).map_err(ParseError::from),
+            FloatDefinition::CoordMP => {
+                write_bit_coord_mp(val, stream, false, false).map_err(ParseError::from)
+            }
+            FloatDefinition::CoordMPLowPrecision => {
+                write_bit_coord_mp(val, stream, false, true).map_err(ParseError::from)
+            }
+            FloatDefinition::CoordMPIntegral => {
+                write_bit_coord_mp(val, stream, true, false).map_err(ParseError::from)
+            }
+            FloatDefinition::FloatNoScale => val.write(stream).map_err(ParseError::from),
+            FloatDefinition::NormalVarFloat => {
+                write_bit_normal(val, stream).map_err(ParseError::from)
+            }
+            FloatDefinition::Scaled {
+                bit_count,
+                low,
+                high,
+            } => {
+                let percentage = (val - low) / (high - low);
+                let scale = float_scale(*bit_count);
+                let raw = (percentage * scale).round() as u32;
+                raw.write_sized(stream, *bit_count as usize)?;
+
+                Ok(())
+            }
+        }
+    }
+}
+
+#[test]
+fn test_send_prop_value_roundtrip() {
+    use bitbuffer::{BitReadBuffer, BitReadStream};
+
+    fn send_prop_value_roundtrip(val: SendPropValue, def: SendPropParseDefinition) {
+        let mut data = Vec::new();
+        let pos = {
+            let mut write = BitWriteStream::new(&mut data, LittleEndian);
+            val.encode(&mut write, &def).unwrap();
+            write.bit_len()
+        };
+        let mut read = BitReadStream::new(BitReadBuffer::new(&data, LittleEndian));
+        assert_eq!(val, SendPropValue::parse(&mut read, &def).unwrap());
+        assert_eq!(pos, read.pos());
+    }
+    send_prop_value_roundtrip(
+        SendPropValue::Integer(0),
+        SendPropParseDefinition::UnsignedInt {
+            changes_often: false,
+            bit_count: 5,
+        },
+    );
+    send_prop_value_roundtrip(
+        SendPropValue::Integer(12),
+        SendPropParseDefinition::NormalVarInt {
+            changes_often: false,
+            unsigned: false,
+        },
+    );
+    send_prop_value_roundtrip(
+        SendPropValue::Integer(12),
+        SendPropParseDefinition::NormalVarInt {
+            changes_often: false,
+            unsigned: false,
+        },
+    );
+    send_prop_value_roundtrip(
+        SendPropValue::Integer(-12),
+        SendPropParseDefinition::NormalVarInt {
+            changes_often: false,
+            unsigned: true,
+        },
+    );
+    send_prop_value_roundtrip(
+        SendPropValue::String("foobar".into()),
+        SendPropParseDefinition::String {
+            changes_often: false,
+        },
+    );
+    send_prop_value_roundtrip(
+        SendPropValue::Vector(Vector {
+            x: 1.0,
+            y: 0.0,
+            z: 1.125,
+        }),
+        SendPropParseDefinition::Vector {
+            changes_often: false,
+            definition: FloatDefinition::Coord,
+        },
+    );
+    send_prop_value_roundtrip(
+        SendPropValue::VectorXY(VectorXY { x: 1.0, y: 0.0 }),
+        SendPropParseDefinition::VectorXY {
+            changes_often: false,
+            definition: FloatDefinition::FloatNoScale,
+        },
+    );
+    send_prop_value_roundtrip(
+        SendPropValue::Float(12.5),
+        SendPropParseDefinition::Float {
+            changes_often: false,
+            definition: FloatDefinition::CoordMP,
+        },
+    );
+    send_prop_value_roundtrip(
+        SendPropValue::Float(12.0),
+        SendPropParseDefinition::Float {
+            changes_often: false,
+            definition: FloatDefinition::CoordMPIntegral,
+        },
+    );
+    send_prop_value_roundtrip(
+        SendPropValue::Float(12.5),
+        SendPropParseDefinition::Float {
+            changes_often: false,
+            definition: FloatDefinition::CoordMPLowPrecision,
+        },
+    );
+    send_prop_value_roundtrip(
+        SendPropValue::Float(12.498169),
+        SendPropParseDefinition::Float {
+            changes_often: false,
+            definition: FloatDefinition::Scaled {
+                bit_count: 12,
+                high: 25.0,
+                low: 10.0,
+            },
+        },
+    );
+    send_prop_value_roundtrip(
+        SendPropValue::Array(vec![
+            SendPropValue::Integer(0),
+            SendPropValue::Integer(1),
+            SendPropValue::Integer(2),
+        ]),
+        SendPropParseDefinition::Array {
+            changes_often: false,
+            inner_definition: Box::new(SendPropParseDefinition::UnsignedInt {
+                changes_often: false,
+                bit_count: 3,
+            }),
+            count_bit_count: 5,
+        },
+    );
+
+    send_prop_value_roundtrip(
+        SendPropValue::Float(76.22549),
+        SendPropParseDefinition::Float {
+            changes_often: false,
+            definition: FloatDefinition::Scaled {
+                bit_count: 10,
+                high: 102.3,
+                low: 0.09990235,
+            },
+        },
+    );
+    send_prop_value_roundtrip(
+        SendPropValue::Vector(Vector {
+            x: 1.0,
+            y: -25.96875,
+            z: 0.1875,
+        }),
+        SendPropParseDefinition::Vector {
+            changes_often: false,
+            definition: FloatDefinition::CoordMP,
+        },
+    );
+    send_prop_value_roundtrip(
+        SendPropValue::Integer(-1),
+        SendPropParseDefinition::NormalVarInt {
+            changes_often: false,
+            unsigned: false,
+        },
+    );
 }
 
 impl From<i32> for SendPropValue {
@@ -702,61 +1009,61 @@ impl From<Vec<SendPropValue>> for SendPropValue {
 }
 
 impl TryFrom<&SendPropValue> for i64 {
-    type Error = ();
+    type Error = MalformedSendPropDefinitionError;
     fn try_from(value: &SendPropValue) -> std::result::Result<Self, Self::Error> {
         match value {
             SendPropValue::Integer(val) => Ok(*val),
-            _ => Err(()),
+            _ => Err(MalformedSendPropDefinitionError::WrongPropType),
         }
     }
 }
 
 impl TryFrom<&SendPropValue> for Vector {
-    type Error = ();
+    type Error = MalformedSendPropDefinitionError;
     fn try_from(value: &SendPropValue) -> std::result::Result<Self, Self::Error> {
         match value {
             SendPropValue::Vector(val) => Ok(*val),
-            _ => Err(()),
+            _ => Err(MalformedSendPropDefinitionError::WrongPropType),
         }
     }
 }
 
 impl TryFrom<&SendPropValue> for VectorXY {
-    type Error = ();
+    type Error = MalformedSendPropDefinitionError;
     fn try_from(value: &SendPropValue) -> std::result::Result<Self, Self::Error> {
         match value {
             SendPropValue::VectorXY(val) => Ok(*val),
-            _ => Err(()),
+            _ => Err(MalformedSendPropDefinitionError::WrongPropType),
         }
     }
 }
 
 impl TryFrom<&SendPropValue> for f32 {
-    type Error = ();
+    type Error = MalformedSendPropDefinitionError;
     fn try_from(value: &SendPropValue) -> std::result::Result<Self, Self::Error> {
         match value {
             SendPropValue::Float(val) => Ok(*val),
-            _ => Err(()),
+            _ => Err(MalformedSendPropDefinitionError::WrongPropType),
         }
     }
 }
 
 impl<'a> TryFrom<&'a SendPropValue> for &'a str {
-    type Error = ();
+    type Error = MalformedSendPropDefinitionError;
     fn try_from(value: &'a SendPropValue) -> std::result::Result<Self, Self::Error> {
         match value {
             SendPropValue::String(val) => Ok(val.as_str()),
-            _ => Err(()),
+            _ => Err(MalformedSendPropDefinitionError::WrongPropType),
         }
     }
 }
 
 impl<'a> TryFrom<&'a SendPropValue> for &'a [SendPropValue] {
-    type Error = ();
+    type Error = MalformedSendPropDefinitionError;
     fn try_from(value: &'a SendPropValue) -> std::result::Result<Self, Self::Error> {
         match value {
             SendPropValue::Array(val) => Ok(val.as_slice()),
-            _ => Err(()),
+            _ => Err(MalformedSendPropDefinitionError::WrongPropType),
         }
     }
 }
@@ -771,32 +1078,84 @@ impl SendPropIdentifier {
     }
 }
 
-#[derive(Debug, Clone, Display)]
+impl From<u64> for SendPropIdentifier {
+    fn from(raw: u64) -> Self {
+        SendPropIdentifier(raw)
+    }
+}
+
+#[derive(Debug, Clone, Display, PartialEq)]
 #[display("{index} = {value}")]
 pub struct SendProp {
-    pub index: SendPropIdentifier,
+    pub index: u32,
+    pub identifier: SendPropIdentifier,
     pub value: SendPropValue,
 }
 
 pub fn read_var_int(stream: &mut Stream, signed: bool) -> ReadResult<i32> {
-    let mut result: i32 = 0;
-
-    for i in (0..35).step_by(7) {
-        let byte: u8 = stream.read()?;
-        result |= ((byte & 0x7F) as i32) << i;
-
-        if (byte >> 7) == 0 {
-            break;
-        }
-    }
+    let abs_int = crate::demo::message::stringtable::read_var_int(stream)? as i32;
 
     if signed {
-        Ok((result >> 1) ^ -(result & 1))
+        Ok((abs_int >> 1) ^ -(abs_int & 1))
     } else {
-        Ok(result)
+        Ok(abs_int)
     }
 }
 
+pub fn write_var_int(
+    int: i32,
+    stream: &mut BitWriteStream<LittleEndian>,
+    signed: bool,
+) -> ReadResult<()> {
+    let abs = if signed {
+        let int = (int << 1) ^ (int >> 31);
+        u32::from_le_bytes(int.to_le_bytes())
+    } else {
+        int as u32
+    };
+
+    crate::demo::message::stringtable::write_var_int(abs, stream)
+}
+
+#[test]
+fn test_var_int_roundtrip() {
+    use bitbuffer::{BitReadBuffer, BitReadStream};
+
+    fn var_int_roundtrip(int: i32, signed: bool) {
+        let mut data = Vec::new();
+        let pos = {
+            let mut write = BitWriteStream::new(&mut data, LittleEndian);
+            write_var_int(int, &mut write, signed).unwrap();
+            write.bit_len()
+        };
+        let mut read = BitReadStream::new(BitReadBuffer::new(&data, LittleEndian));
+        assert_eq!(int, read_var_int(&mut read, signed).unwrap());
+        assert_eq!(pos, read.pos());
+    }
+    var_int_roundtrip(0, false);
+    var_int_roundtrip(1, false);
+    var_int_roundtrip(10, false);
+    var_int_roundtrip(55, false);
+    var_int_roundtrip(355, false);
+    var_int_roundtrip(12354, false);
+    var_int_roundtrip(123125412, false);
+
+    var_int_roundtrip(0, true);
+    var_int_roundtrip(1, true);
+    var_int_roundtrip(10, true);
+    var_int_roundtrip(55, true);
+    var_int_roundtrip(355, true);
+    var_int_roundtrip(12354, true);
+    var_int_roundtrip(123125412, true);
+    var_int_roundtrip(-0, true);
+    var_int_roundtrip(-1, true);
+    var_int_roundtrip(-10, true);
+    var_int_roundtrip(-55, true);
+    var_int_roundtrip(-355, true);
+    var_int_roundtrip(-12354, true);
+    var_int_roundtrip(-123125412, true);
+}
+
 pub fn read_bit_coord(stream: &mut Stream) -> ReadResult<f32> {
     let has_int = stream.read()?;
     let has_frac = stream.read()?;
@@ -816,6 +1175,56 @@ pub fn read_bit_coord(stream: &mut Stream) -> ReadResult<f32> {
     })
 }
 
+pub fn write_bit_coord(val: f32, stream: &mut BitWriteStream<LittleEndian>) -> ReadResult<()> {
+    let has_int = val.abs() >= 1.0;
+    has_int.write(stream)?;
+    let has_frac = val.fract() != 0.0;
+    has_frac.write(stream)?;
+
+    if has_frac || has_int {
+        let sign = val.is_negative();
+        sign.write(stream)?;
+    }
+    let abs = val.abs();
+    if has_int {
+        (abs as u16 - 1).write_sized(stream, 14)?;
+    }
+    if has_frac {
+        let frac_val = (abs.fract() / get_frac_factor(5)) as u8;
+        frac_val.write_sized(stream, 5)?;
+    }
+    Ok(())
+}
+
+#[test]
+fn bit_coord_roundtrip() {
+    use bitbuffer::BitReadBuffer;
+
+    let mut data = Vec::with_capacity(16);
+    let (pos1, pos2, pos3, pos4) = {
+        let mut write = BitWriteStream::new(&mut data, LittleEndian);
+        write_bit_coord(0.0, &mut write).unwrap();
+        let pos1 = write.bit_len();
+        write_bit_coord(123.0, &mut write).unwrap();
+        let pos2 = write.bit_len();
+        write_bit_coord(123.4375, &mut write).unwrap();
+        let pos3 = write.bit_len();
+        write_bit_coord(-0.4375, &mut write).unwrap();
+        let pos4 = write.bit_len();
+        (pos1, pos2, pos3, pos4)
+    };
+
+    let mut read = Stream::from(BitReadBuffer::new(&data, LittleEndian));
+    assert_eq!(0.0, read_bit_coord(&mut read).unwrap());
+    assert_eq!(pos1, read.pos());
+    assert_eq!(123.0, read_bit_coord(&mut read).unwrap());
+    assert_eq!(pos2, read.pos());
+    assert_eq!(123.4375, read_bit_coord(&mut read).unwrap());
+    assert_eq!(pos3, read.pos());
+    assert_eq!(-0.4375, read_bit_coord(&mut read).unwrap());
+    assert_eq!(pos4, read.pos());
+}
+
 fn get_frac_factor(bits: usize) -> f32 {
     1.0 / ((1 << bits) as f32)
 }
@@ -856,6 +1265,77 @@ pub fn read_bit_coord_mp(
     Ok(value)
 }
 
+pub fn write_bit_coord_mp(
+    val: f32,
+    stream: &mut BitWriteStream<LittleEndian>,
+    is_integral: bool,
+    low_precision: bool,
+) -> ReadResult<()> {
+    let abs = val.abs();
+    let in_bounds = (abs as u32) < (1 << 11);
+    let has_int_val = abs >= 1.0;
+    in_bounds.write(stream)?;
+    has_int_val.write(stream)?;
+
+    if is_integral {
+        if has_int_val {
+            val.is_sign_negative().write(stream)?;
+            ((abs - 1.0) as u32).write_sized(stream, if in_bounds { 11 } else { 14 })?;
+        }
+    } else {
+        val.is_sign_negative().write(stream)?;
+        if has_int_val {
+            ((abs - 1.0) as u32).write_sized(stream, if in_bounds { 11 } else { 14 })?;
+        }
+        let frac_bits = if low_precision { 3 } else { 5 };
+        let frac_val = (abs.fract() / get_frac_factor(frac_bits)) as u32;
+        frac_val.write_sized(stream, frac_bits)?;
+    }
+
+    Ok(())
+}
+
+#[test]
+fn test_bit_coord_mp_roundtrip() {
+    use bitbuffer::{BitReadBuffer, BitReadStream};
+
+    fn bit_coord_mp_normal(val: f32, is_integral: bool, low_precision: bool) {
+        let mut data = Vec::with_capacity(16);
+        let pos = {
+            let mut write = BitWriteStream::new(&mut data, LittleEndian);
+            write_bit_coord_mp(val, &mut write, is_integral, low_precision).unwrap();
+            write.bit_len()
+        };
+        let mut read = BitReadStream::new(BitReadBuffer::new(&data, LittleEndian));
+        assert_eq!(
+            val,
+            read_bit_coord_mp(&mut read, is_integral, low_precision).unwrap()
+        );
+        assert_eq!(pos, read.pos());
+    }
+
+    bit_coord_mp_normal(1.0, false, false);
+
+    bit_coord_mp_normal(0.0, false, false);
+    bit_coord_mp_normal(0.5, false, false);
+    bit_coord_mp_normal(-0.5, false, false);
+    bit_coord_mp_normal(1234.5, false, false);
+    bit_coord_mp_normal(-1234.5, false, false);
+    bit_coord_mp_normal(2.0f32.powf(12.0) + 0.125, false, false);
+
+    bit_coord_mp_normal(0.0, false, true);
+    bit_coord_mp_normal(0.5, false, true);
+    bit_coord_mp_normal(-0.5, false, true);
+    bit_coord_mp_normal(1234.5, false, true);
+    bit_coord_mp_normal(-1234.5, false, true);
+    bit_coord_mp_normal(2.0f32.powf(12.0) + 0.125, false, true);
+
+    bit_coord_mp_normal(0.0, true, false);
+    bit_coord_mp_normal(1234.0, true, false);
+    bit_coord_mp_normal(-1234.0, true, false);
+    bit_coord_mp_normal(2.0f32.powf(12.0), true, false);
+}
+
 pub fn read_bit_normal(stream: &mut Stream) -> ReadResult<f32> {
     let is_negative = stream.read()?;
     let frac_val: u16 = stream.read_sized(11)?;
@@ -866,3 +1346,30 @@ pub fn read_bit_normal(stream: &mut Stream) -> ReadResult<f32> {
         Ok(value)
     }
 }
+
+pub fn write_bit_normal(val: f32, stream: &mut BitWriteStream<LittleEndian>) -> ReadResult<()> {
+    val.is_sign_negative().write(stream)?;
+    let frac_val = (val.abs().fract() / get_frac_factor(11)) as u16;
+    frac_val.write_sized(stream, 11)
+}
+
+#[test]
+fn test_bit_normal_roundtrip() {
+    use bitbuffer::{BitReadBuffer, BitReadStream};
+
+    fn roundtrip_normal(val: f32) {
+        let mut data = Vec::with_capacity(16);
+        let pos = {
+            let mut write = BitWriteStream::new(&mut data, LittleEndian);
+            write_bit_normal(val, &mut write).unwrap();
+            write.bit_len()
+        };
+        let mut read = BitReadStream::new(BitReadBuffer::new(&data, LittleEndian));
+        assert_eq!(val, read_bit_normal(&mut read).unwrap());
+        assert_eq!(pos, read.pos());
+    }
+    roundtrip_normal(0.0);
+    roundtrip_normal(-0.0);
+    roundtrip_normal(0.5);
+    roundtrip_normal(-0.5);
+}
